from .config import TypedCoreConfig

from ._state_machine import DEPLOYMENT_READY_CONDITIONS
from .app_config import AppConfig, AppConfigError
from .capsule import CapsuleDeployer, list_and_filter_capsules
from functools import partial
import sys
from typing import Type


class AppDeployer(TypedCoreConfig):
    """ """

    __init__ = TypedCoreConfig.__init__

    _app_config: AppConfig

    _state = {}

    @property
    def app_config(self) -> AppConfig:
        if not hasattr(self, "_app_config"):
            self._app_config = AppConfig(self._config)
        return self._app_config

    # Things that need to be set before deploy
    @classmethod
    def _set_state(
        cls,
        perimeter: str,
        api_url: str,
        code_package_url: str = None,
        code_package_key: str = None,
        name: str = None,
        image: str = None,
    ):
        cls._state["perimeter"] = perimeter
        cls._state["api_url"] = api_url
        cls._state["code_package_url"] = code_package_url
        cls._state["code_package_key"] = code_package_key
        cls._state["name"] = name
        cls._state["image"] = image

    def deploy(
        self,
        readiness_condition=DEPLOYMENT_READY_CONDITIONS.ATLEAST_ONE_RUNNING,
        max_wait_time=600,
        readiness_wait_time=10,
        logger_fn=partial(print, file=sys.stderr),
        status_file=None,
        no_loader=False,
        **kwargs,
    ):
        # Name setting from top level if none is set in the code
        if self.app_config._core_config.name is None:
            self.app_config._core_config.name = self._state["name"]

        self.app_config.commit()

        # Set any state that might have been passed down from the top level
        for k, v in self._state.items():
            if self.app_config.get_state(k) is None:
                self.app_config.set_state(k, v)

        capsule = CapsuleDeployer(
            self.app_config,
            self._state["api_url"],
            create_timeout=max_wait_time,
            debug_dir=None,
            success_terminal_state_condition=readiness_condition,
            readiness_wait_time=readiness_wait_time,
            logger_fn=logger_fn,
        )

        currently_present_capsules = list_and_filter_capsules(
            capsule.capsule_api,
            None,
            None,
            capsule.name,
            None,
            None,
            None,
        )

        force_upgrade = self.app_config.get_state("force_upgrade", False)

        if len(currently_present_capsules) > 0:
            # Only update the capsule if there is no upgrade in progress
            # Only update a "already updating" capsule if the `--force-upgrade` flag is provided.
            _curr_cap = currently_present_capsules[0]
            this_capsule_is_being_updated = _curr_cap.get("status", {}).get(
                "updateInProgress", False
            )

            if this_capsule_is_being_updated and not force_upgrade:
                _upgrader = _curr_cap.get("metadata", {}).get("lastModifiedBy", None)
                message = f"{capsule.capsule_type} is currently being upgraded"
                if _upgrader:
                    message = (
                        f"{capsule.capsule_type} is currently being upgraded. Upgrade was launched by {_upgrader}. "
                        "If you wish to force upgrade, you can do so by providing the `--force-upgrade` flag."
                    )
                raise AppConfigError(message)

            logger_fn(
                f"🚀 {'' if not force_upgrade else 'Force'} Upgrading {capsule.capsule_type.lower()} `{capsule.name}`....",
            )
        else:
            logger_fn(
                f"🚀 Deploying {capsule.capsule_type.lower()} `{capsule.name}`....",
            )

        capsule.create()
        final_status = capsule.wait_for_terminal_state()
        return final_status


class apps:

    _name_prefix = None

    @classmethod
    def set_name_prefix(cls, name_prefix: str):
        cls._name_prefix = name_prefix

    @property
    def name_prefix(self) -> str:
        return self._name_prefix

    @property
    def Deployer(self) -> Type[AppDeployer]:
        return AppDeployer



from metaflow.exception import MetaflowException
from metaflow.decorators import StepDecorator
from metaflow import current
from .core import AppDeployer, apps
from .core.perimeters import PerimeterExtractor
import os
import hashlib


class AppDeployDecorator(StepDecorator):

    """
    MF Add To Current
    -----------------
    apps -> metaflow_extensions.outerbounds.plugins.apps.core.apps

        @@ Returns
        ----------
        apps
            The object carrying the Deployer class to deploy apps.
    """

    name = "app_deploy"
    defaults = {}

    package_url = None
    package_sha = None

    def step_init(self, flow, graph, step, decos, environment, flow_datastore, logger):
        self.logger = logger
        self.environment = environment
        self.step = step
        self.flow_datastore = flow_datastore

    def _resolve_package_url_and_sha(self):
        return os.environ.get("METAFLOW_CODE_URL", self.package_url), os.environ.get(
            "METAFLOW_CODE_SHA", self.package_sha
        )

    def task_pre_step(
        self,
        step_name,
        task_datastore,
        metadata,
        run_id,
        task_id,
        flow,
        graph,
        retry_count,
        max_user_code_retries,
        ubf_context,
        inputs,
    ):
        perimeter, api_server = PerimeterExtractor.during_metaflow_execution()
        package_url, package_sha = self._resolve_package_url_and_sha()
        if package_url is None or package_sha is None:
            raise MetaflowException(
                "METAFLOW_CODE_URL or METAFLOW_CODE_SHA is not set. "
                "Please set METAFLOW_CODE_URL and METAFLOW_CODE_SHA in your environment."
            )
        default_name = "-".join(current.pathspec.split("/")).lower()
        image = os.environ.get("FASTBAKERY_IMAGE", None)

        hash_key = hashlib.sha256(package_url.encode()).hexdigest()[:6]

        default_name = (
            (current.flow_name + "-" + current.step_name)[:12] + "-" + hash_key
        ).lower()

        AppDeployer._set_state(
            perimeter,
            api_server,
            code_package_url=package_url,
            code_package_key=package_sha,
            name=default_name,
            image=image,
        )
        current._update_env(
            {
                "apps": apps(),
            }
        )

    def task_post_step(
        self, step_name, flow, graph, retry_count, max_user_code_retries
    ):
        pass

    def runtime_init(self, flow, graph, package, run_id):
        # Set some more internal state.
        self.flow = flow
        self.graph = graph
        self.package = package
        self.run_id = run_id

    def runtime_task_created(
        self, task_datastore, task_id, split_index, input_paths, is_cloned, ubf_context
    ):
        # To execute the Kubernetes job, the job container needs to have
        # access to the code package. We store the package in the datastore
        # which the pod is able to download as part of it's entrypoint.
        if not is_cloned:
            self._save_package_once(self.flow_datastore, self.package)

    @classmethod
    def _save_package_once(cls, flow_datastore, package):
        if cls.package_url is None:
            cls.package_url, cls.package_sha = flow_datastore.save_data(
                [package.blob], len_hint=1
            )[0]
            os.environ["METAFLOW_CODE_URL"] = cls.package_url
            os.environ["METAFLOW_CODE_SHA"] = cls.package_sha
